import dash.dcc as dcc
import dash.html as html
import dash_bootstrap_components as dbc
import numpy as np
from dash import (
    Dash,
    Input,
    Output,
    Patch,
    State,
    callback,
)
from omegaconf import OmegaConf

from ..footprint.find import simple_peaks
from ..main import (
    data,
    find,
    get_frame,
    reduce,
    stats,
)
from .fig import (
    bar_fig,
    circle_fig,
    heat_fig,
    scatter_fig,
    update_img,
)
from .ui import (
    Collapse,
    ThreadButton,
    two_column,
)


def HotaruApp(cfg, *args, **kwargs):
    kwargs.setdefault("external_stylesheets", [dbc.themes.BOOTSTRAP])
    kwargs.setdefault("title", "HOTARU")
    app = Dash(__name__, *args, **kwargs)

    divs = [
        cfg_store := dcc.Store("cfg_store", data=OmegaConf.to_container(cfg)),
    ]

    load_div = Collapse(
        "load",
        True,
        pbar := dbc.Progress(),
        html.Div(
            style=two_column(500),
            children=[
                dbc.Label("Image file path"),
                imgs_path := dbc.Input(type="text", value=cfg.data.imgs.file),
                dbc.Label("Mask type"),
                imgs_mask := dbc.Input(type="text", value=cfg.data.mask.type),
                dbc.Label("Frequency"),
                imgs_hz := dbc.Input(type="number", value=cfg.data.hz),
                dbc.Label("Progress"),
            ],
        ),
        load_btn := ThreadButton("STATS", stats, cfg_store, pbar),
        html.Div(
            style=two_column(1200),
            children=[
                dbc.Label("Radius range"),
                radius := dcc.RangeSlider(
                    min=0,
                    max=5,
                    step=0.1,
                    value=[1, 4],
                    marks={i: str(2**i) for i in range(6)},
                ),
                dbc.Label("Num radius"),
                nradius := dcc.Slider(min=1, max=50, step=1, value=11),
                dbc.Label("Radius cut range"),
                radius2 := dcc.RangeSlider(
                    id="radius-range2",
                    min=0,
                    max=10,
                    step=1,
                    value=[1, 9],
                    marks={
                        i: f"{2 ** r:.2f}" for i, r in enumerate(np.linspace(1, 5, 11))
                    },
                ),
                dbc.Label("Distance threshold", html_for="thr"),
                thr := dcc.Slider(id="thr", min=0.5, max=3, step=0.1, value=2.0),
            ],
        ),
    )

    stats_div = Collapse(
        "stats",
        False,
        html.H2("Conventional cell candidates detection"),
        html.Div(
            style=two_column(1200),
            children=[
                dbc.Label("Gaussian"),
                gaussian := dcc.Slider(
                    min=0,
                    max=10,
                    step=0.1,
                    value=1,
                    marks={i: str(i) for i in range(11)},
                ),
                dbc.Label("MaxPool"),
                maxpool := dcc.Slider(min=3, max=11, step=2, value=3),
            ],
        ),
        stats_graph := dbc.Row(
            children=[
                dbc.Col(std_graph := dcc.Graph(figure=heat_fig("std")), width="auto"),
                dbc.Col(cor_graph := dcc.Graph(figure=heat_fig("cor")), width="auto"),
                dbc.Col(
                    std_cor_graph := dcc.Graph(
                        figure=scatter_fig("std-cor", "std", "cor")
                    ),
                    width="auto",
                ),
            ],
        ),
    )

    divs += [load_div, stats_div]

    frame_div = Collapse(
        "frame",
        False,
        html.H2("Check frames"),
        html.Div(
            style=two_column(1200),
            children=[
                dbc.Label("Frame"),
                frame := dcc.Slider(
                    updatemode="drag",
                    min=0,
                    max=0,
                    step=1,
                    value=0,
                ),
                dbc.Label("Min/Max"),
                minmax := dcc.RangeSlider(
                    min=-1,
                    max=1,
                    step=0.1,
                    value=[-1, 1],
                ),
            ],
        ),
        frame_graph := dbc.Row(
            children=[
                dbc.Col(norm_img := dcc.Graph(figure=heat_fig("norm")), width="auto"),
                dbc.Col(
                    hist_img := dcc.Graph(
                        figure=bar_fig("intensity histogram", "intensity", "count")
                    ),
                    width="auto",
                ),
            ],
        ),
    )

    divs += [frame_div]

    @callback(
        Output(std_graph, "figure"),
        Output(cor_graph, "figure"),
        Output(std_cor_graph, "figure"),
        load_btn.finish,
        Input(gaussian, "value"),
        Input(maxpool, "value"),
        State(cfg_store, "data"),
        prevent_initial_call=True,
    )
    def plot_stats(finish, gaussian, maxpool, cfg):
        cfg = OmegaConf.create(cfg)
        imin, imax, istd, icor = stats(cfg)
        y, x = simple_peaks(icor, gaussian, maxpool)
        std_fig = Patch()
        std_fig.data[0].z = istd
        std_fig.data[0].zmin = istd.min()
        std_fig.data[0].zmax = istd.max()
        std_fig.data[1].x = x
        std_fig.data[1].y = y
        cor_fig = Patch()
        cor_fig.data[0].z = icor
        cor_fig.data[0].zmin = icor.min()
        cor_fig.data[0].zmax = icor.max()
        cor_fig.data[1].x = x
        cor_fig.data[1].y = y
        scatter = Patch()
        scatter.data[0].x = istd[y, x]
        scatter.data[0].y = icor[y, x]
        return (
            update_img(std_fig, *istd.shape),
            update_img(cor_fig, *icor.shape),
            scatter,
        )

    @callback(
        Output(frame, "max"),
        Output(frame, "marks"),
        Output(frame, "value"),
        Output(minmax, "min"),
        Output(minmax, "max"),
        Output(minmax, "marks"),
        load_btn.finish,
        State(cfg_store, "data"),
        prevent_initial_call=True,
    )
    def update_frame(finish, cfg):
        cfg = OmegaConf.create(cfg)
        nt = data(cfg).imgs.shape[0]
        imin, imax, istd, icor = stats(cfg)
        fmarks = {i: f"{i}" for i in range(0, nt + 1, nt // 10)}
        m = np.floor(10 * imin.min()) / 10
        M = np.ceil(10 * imax.max()) / 10
        mmarks = {float(v): f"{v:.1f}" for v in np.arange(m, M, 1)}
        return nt - 1, fmarks, 0, m, M, mmarks

    @callback(
        Output(norm_img, "figure"),
        Output(hist_img, "figure"),
        load_btn.finish,
        Input(frame, "value"),
        Input(minmax, "value"),
        State(cfg_store, "data"),
        prevent_initial_call=True,
    )
    def plot_norm(finish, t, minmax, cfg):
        cfg = OmegaConf.create(cfg)
        vmin, vmax = minmax
        img = get_frame(cfg, t)
        norm_fig = Patch()
        norm_fig.data[0].z = img
        norm_fig.data[0].zmin = vmin
        norm_fig.data[0].zmax = vmax
        gcount, ghist = np.histogram(img.ravel(), bins=np.linspace(-0.1, 5.0, 100))
        hist_fig = Patch()
        hist_fig.data[0].x = ghist
        hist_fig.data[0].y = gcount
        return update_img(norm_fig, *img.shape), hist_fig

    Collapse(
        "peak",
        True,
        html.H2("Find peaks"),
        find_btn := ThreadButton("PEAK", find, pbar),
        dbc.Row(
            children=[
                dcc.Store("peak"),
                dbc.Col(
                    circle := dcc.Graph(figure=circle_fig("cell candidates")),
                    width="auto",
                ),
                dbc.Col(
                    radius_intensity := dcc.Graph(
                        figure=scatter_fig("radius-intensity", "radius", "intensity")
                    ),
                    width="auto",
                ),
            ],
        ),
    )

    @callback(
        Output(circle, "figure"),
        Output(radius_intensity, "figure"),
        find_btn.finish,
        State(cfg_store, "data"),
        prevent_initial_call=True,
    )
    def plot_circle(finish, cfg):
        peaks = reduce(cfg)
        fig1 = Patch()
        fig1.data[0].x = peaks.x
        fig1.data[0].y = peaks.y
        fig1.data[0].marker.opacity = 0.5 * peaks.v / peaks.v.max()
        fig1.data[0].marker.size = peaks.r
        fig1.data[0].marker.sizeref = 0.25 * h / 500
        fig1.data[0].marker.sizemode = "diameter"

        find = find(cfg)
        dr = find.radius[1] / find.radius[0] - 1
        v0 = find.v.ravel()
        r0 = find.r.ravel()
        jitter0 = 1 + 0.2 * dr * np.random.randn(v0.size)
        jitter = 1 + 0.2 * dr * np.random.randn(reduce.r.size)
        fig2 = Patch()
        fig2.data[0].x = jitter0 * r0
        fig2.data[0].y = v0
        fig2.data[0].marker.maxdisplayed = 10000
        fig2.data[0].marker.opacity = 0.3
        fig2.data[1].x = jitter * peaks.r
        fig2.data[1].y = peaks.v
        fig2.data[1].marker.opacity = 0.3
        fig2.layout.xaxis.type = "log"
        return update_img(fig1, *shape), fig2

    """
    filt_div = Collapse(
        "filt",
        False,
        filt_btn := ThreadButton("TEST", filt, pbar),
        filt_graph := dbc.Row(
            children=[
                dbc.Col(filt_img := dcc.Graph(figure=heat_fig("filt", "Picnic", "green", -0.6, 0.6))),
            ],
        ),
    )

    @callback(
        Output(filt_img, "figure"),
        filt_btn.finish,
        prevent_initial_call=True,
    )
    def plot_filt(data):
        log_img = find_peaks(cur_img[None, ...], mask, radius)
        gl, t, y, x, r, v = cache["log_img"]
        nt, nr, h, w = gl.shape
        log_img = np.pad(gl[0, ...], [[0, 0], [0, 0], [2, 2]])
        log_img = np.concatenate(log_img, axis=1)
        fig = Patch()
        fig.data[0].z = log_img
        fig.data[1].x = 2 + (w + 4) * r + x
        fig.data[1].y = y
        return update_img(fig, *log_img.shape, 200)

    make_div = Collapse(
        "make",
        True,
        html.H2("Initial step"),
        make := ThreadButton(
            "FOOTPRINT",
            lambda: None,
            model().segments,
        ),
        spike := ThreadButton(
            "SPIKE",
            lambda: None,
            lambda pbar: model().spikes(True, pbar),
        ),
        dbc.Row(
            children=[
                dbc.Col(seg_img := dcc.Graph(figure=heat_fig("initial footprints")), width="auto"),
                dbc.Col(spk_img := dcc.Graph(figure=spike_fig("initial spikes", cache["cfg"].data.hz)), width="auto"),
            ],
        ),
    )

    @callback(
        Output(seg_img, "figure"),
        make.finish,
        prevent_initial_call=True,
    )
    def plot_seg(data):
        seg = model().segments().max(axis=0)
        fig = Patch()
        fig.data[0].z = seg
        return update_img(fig, *seg.shape)

    @callback(
        Output(spk_img, "figure"),
        spike.finish,
        prevent_initial_call=True,
    )
    def plot_spk(data):
        spk = model().spikes(True)
        print(spk.shape)
        #spk /= spk.max(axis=0, keepdims=True)
        fig = Patch()
        fig.data[0].z = spk[0]
        return fig

    divs += [find_div, make_div]
    """

    @callback(
        Output(cfg_store, "data"),
        Output(radius2, "max"),
        Output(radius2, "marks"),
        Input(imgs_path, "value"),
        Input(imgs_mask, "value"),
        Input(imgs_hz, "value"),
        Input(radius, "value"),
        Input(nradius, "value"),
        Input(radius2, "value"),
        Input(thr, "value"),
        State(cfg_store, "data"),
    )
    def set_cfg(imgs, mask, hz, radius1, rnum, radius2, thr, cfg):
        rmin, rmax = radius
        rmin2, rmax2 = radius2
        radius = np.geomspace(rmin, rmax, rnum)
        marks = {i: f"{r:.2f}" for i, r in enumerate(radius)}
        cfg = OmegaConf.create(cfg)
        cfg.data.imgs.file = imgs
        cfg.data.mask.type = mask
        cfg.data.hz = hz
        cfg.radius.rmin = rmin
        cfg.radius.rmax = rmax
        cfg.radius.rnum = rnum
        cfg.radius.rmin2 = rmin2
        cfg.radius.rmax2 = rmax2
        cfg.radius.thr = thr
        return OmegaConf.to_container(cfg), rnum - 1, marks

    app.layout = html.Div(divs)
    return app



import numpy as np
import plotly.graph_objects as go
import plotly.io as pio

pio.kaleido.scope.mathjax = None


def plot_data(paths, imgs, labels, dlabels, width, margin, pad):

    num = len(imgs)

    hs, ws = zip(*(np.load(f"{path}/{imgs[0]}.npy").shape for path in paths))
    print(hs, ws)
    aspects = np.array(hs) / np.array(ws)

    fig = go.Figure().set_subplots(
        len(paths),
        len(imgs),
        column_titles=labels,
        row_titles=dlabels,
        row_heights=list(aspects),
        horizontal_spacing=0.01,
        vertical_spacing=0.03,
    )
    for i, path in enumerate(paths):
        for j, img in enumerate(imgs):
            img = np.load(f"{path}/{img}.npy")
            h, w = img.shape
            print(h, w)
            smin = img.min()
            smax = img.max()
            img = (img - smin) / (smax - smin)
            fig.add_trace(
                go.Heatmap(z=img, colorscale="Greens", showscale=False),
                row=i + 1,
                col=j + 1,
            )
            fig.update_xaxes(
                autorange=False,
                range=(0, w),
                showticklabels=False,
                row=i + 1,
                col=j + 1,
            )
            fig.update_yaxes(
                # scaleanchor="x",
                autorange=False,
                range=(h, 0),
                showticklabels=False,
                row=i + 1,
                col=j + 1,
            )
    mod_width = width - margin["l"] - margin["r"]
    fig.update_layout(
        annotations=[dict(font=dict(size=11)) for _ in range(num)],
        width=width,
        height=np.ceil(margin["t"] + margin["b"] + mod_width * aspects.sum() / num),
        margin=margin,
    )
    return fig



import jax.lax as lax
import jax.numpy as jnp

__all__ = [
    "neighbor",
]


def neighbor(imgs):
    kernel = jnp.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]], jnp.float32) / 8
    kernel = jnp.expand_dims(kernel, (0, 1))
    return lax.conv(jnp.expand_dims(imgs, -3), kernel, (1, 1), "same")[..., 0, :, :]



import os
import math
import subprocess
import psutil
from logging import getLogger

import tensorflow as tf
import jax
from jax.experimental.mesh_utils import create_device_mesh
from jax.sharding import PositionalSharding
from jax.dlpack import from_dlpack

logger = getLogger(__name__)


def from_tf(x_tf):
    x_dl = tf.experimental.dlpack.to_dlpack(x_tf)
    return from_dlpack(x_dl)


def get_gpu_env(env):
    match env:
        case GpuEnv():
            return env
        case None:
            return GpuEnv()
        case _:
            return GpuEnv(**env)


class GpuEnv:
    def __init__(self, num_devices=-1, memsize=-1, label="gpu"):
        if not (num_devices > 0 and memsize > 0):
            gpu_info = get_gpu_info()
            if not (num_devices > 0):
                num_devices = len(gpu_info)
            if not (memsize > 0):
                memsize = int(min(gpu_info[: num_devices]))
        self.num_devices = num_devices
        self.memsize = memsize * 1e6
        self.label = label

    def batch(self, factor, size):
        n = self.num_devices
        batch_simple = min(size, max(n, int(n * self.memsize / factor)))
        batch_padded = n * ((batch_simple + n - 1) // n)
        return batch_padded

    def batch_sqrt(self, factor, size):
        n = self.num_devices
        batch_size = min(size, max(n, int(n * math.sqrt(self.memsize / factor))))
        batch = n, (batch_size + n - 1) // n
        logger.debug("batch: %s %s", (n, self.memsize, factor, size), batch)
        return batch

    def sharding(self, shape):
        if shape is None:
            shape = (self.num_devices,)
        return PositionalSharding(create_device_mesh(shape))


def get_gpu_info():
    result = subprocess.run(
        [
            "nvidia-smi",
            "--query-gpu=memory.total",
            "--format=csv,noheader,nounits",
        ],
        stdout=subprocess.PIPE,
        encoding="utf-8",
        check=True,
    )
    return [int(x) for x in result.stdout.strip().split(os.linesep)]


def get_gpu_used():
    result = subprocess.run(
        [
            "nvidia-smi",
            "--query-gpu=memory.used",
            "--format=csv,noheader,nounits",
        ],
        stdout=subprocess.PIPE,
        encoding="utf-8",
        check=True,
    )
    return [int(x) for x in result.stdout.strip().split(os.linesep)]


def get_xla_stats():
    backend = jax.lib.xla_bridge.get_backend()
    lbs = backend.live_buffers()
    les = backend.live_executables()
    mem = psutil.Process().memory_info().rss
    return dict(mem=mem * 1e-6, executable=len(les), buffer=[lb.shape for lb in lbs])


def delete_xla_buffers():
    backend = jax.lib.xla_bridge.get_backend()
    for buf in backend.live_buffers():
        buf.delete()
    #for exe in backend.live_executables():
    #    exe.delete()



from .penalty import (
    Penalty,
    get_penalty,
)
from .regularizer import (
    MaxNormNonNegativeL1,
    NonNegativeL1,
    Regularizer,
)
from .model import (
    SpatialModel,
    TemporalModel,
)

__all__ = [
    "Regularizer",
    "NonNegativeL1",
    "MaxNormNonNegativeL1",
    "Penalty",
    "SpikeToCalcium",
    "get_dynamics",
    "get_penalty",
    "SpatialModel",
    "TemporalModel",
]



from logging import getLogger

import numpy as np
import pandas as pd

from ..io import (
    save,
    try_load,
)
from ..train import (
    TemporalModel,
    get_penalty,
)
from ..utils import (
    get_clip,
    get_xla_stats,
)
from ..spike import (
    evaluate,
)
from .common import (
    get_data,
    get_files,
    get_force,
    rev_index,
)

logger = getLogger(__name__)


def temporal(cfg, stage, force=False):
    (statsfile,) = get_files(cfg, "evaluate", stage)
    if (
        get_force(cfg, "evaluate", stage)
        or not statsfile.exists()
    ):
        if stage == 0:
            footprints = try_load(get_files(cfg, "make", stage))
            stats = try_load(get_files(cfg, "init", stage))
        else:
            stats, footprints = try_load(get_files(cfg, "clean", stage))
        spikefile, bgfile, lossfile = get_files(cfg, "temporal", stage)
        if (
            get_force(cfg, "temporal", stage)
            or not spikefile.exists()
            or not bgfile.exists()
        ):
            logger.info(f"exec temporal ({stage})")
            data = get_data(cfg)
            logger.debug("%s", get_xla_stats())
            model = TemporalModel(
                data,
                stats,
                footprints,
                cfg.dynamics,
                get_penalty(cfg.penalty, stage),
            )
            clips = get_clip(data.shape, cfg.cmd.temporal.clip)
            logdfs = []
            out = []
            for i, clip in enumerate(clips):
                model.prepare(clip, **cfg.cmd.temporal.prepare)
                log = model.fit(**cfg.cmd.temporal.step)
                logger.debug("%s", get_xla_stats())

                loss, sigma = zip(*log)
                df = pd.DataFrame(
                    dict(
                        stage=stage,
                        kind="temporal",
                        div=i,
                        step=np.arange(len(log)),
                        loss=loss,
                        sigma=sigma,
                    ),
                )
                logdfs.append(df)
                out.append(model.get_x())
            logdf = pd.concat(logdfs, axis=0)
            index1, index2, x1, x2 = (np.concatenate(v, axis=0) for v in zip(*out))
            spikes = np.array(x1[rev_index(index1)])
            bg = np.array(x2[rev_index(index2)])
            if cfg.fix_top:
                nk, nu = spikes.shape
                idx = np.argpartition(spikes, -2, axis=1)
                spikes[np.arange(nk), idx[:, -1]] = spikes[np.arange(nk), idx[:, -2]]
            save((spikefile, bgfile, lossfile), (spikes, bg, logdf))
            logger.info(f"saved temporal ({stage})")
        else:
            logger.info(f"load temporal ({stage})")
            spikes, bg, logdf = try_load((spikefile, bgfile, lossfile))

        logger.info(f"exec temporal stats ({stage})")
        stats = evaluate(stats, spikes, bg)
        save((statsfile,), (stats,))
        logger.info(f"saved temporal stats ({stage})")



import pandas as pd

from hotaru.plot.complu import compare_lu_fig


paths = [
    "outputs/CA3/miniature/min21/default/u0/000stats.csv",
    "outputs/CA3/miniature/min21/default/u20/000stats.csv",
    "outputs/CA3/miniature/min21/default/u40/000stats.csv",
]
compare_lu_fig(paths, ["u0", "u20", "u40"]).write_image(
    "figs/lu.pdf",
)



from tqdm import tqdm


def get_progress(pbar):
    match pbar:
        case Progress():
            return pbar
        case "silent" | False | None:
            return Progress()
        case "simple":
            return ConsoleProgress()
        case "tqdm":
            return TQDMProgress()


class Progress:
    def skip(self):
        pass

    def session(self, name, total, status=None):
        pass

    def update(self, n, status=None):
        pass

    def close(self):
        pass


class ConsoleProgress(Progress):
    def __init__(self, setter):
        self._setter = setter

    def skip(self):
        pass

    def session(self, name, total, status=None):
        print("session", name)
        self._name = name
        self._setter((self._name, "", "", ""))
        print("set_count", total, status)
        self._status = status
        self._total = total if total > 0 else None
        self._n = 0
        self._setter((self._name, str(self._n), str(self._total), str(self._status)))

    def update(self, n, status=""):
        print("update", n, status)
        self._status = status
        self._n += n
        self._setter((self._name, str(self._n), str(self._total), str(self._status)))


class TQDMProgress(Progress):
    def __init__(self):
        self._curr = None

    def skip(self):
        pass

    def session(self, name, total, status=None):
        self.close()
        self._name = name
        self._curr = tqdm(desc=self._name, total=total)
        if status is not None:
            self._curr.set_postfix_str(status, refresh=False)

    def update(self, n, status=None):
        if status is not None:
            self._curr.set_postfix_str(status, refresh=False)
        self._curr.update(n)

    def close(self):
        if self._curr is not None:
            self._curr.close()
        self._curr = None


"""
class SimpleProgress(Progress):
    def __init__(self, total=None):
        self.reset(total)

    def reset(self, total):
        self.total = total
        self.n = 0

    def update(self, n):
        self.n += n

    @property
    def value(self):
        return 100 * self.n / self.total

    @property
    def label(self):
        return f"{self.n} / {self.total}"
"""



from hotaru.plot.data import plot_data
from hotaru.plot.stats import plot_stats


paths = [
    "outputs/Kd32",
    "outputs/Sato1",
    "outputs/CA3",
]
dlabels = [
    "Data 1", "Data 2", "Data 3",
]
imgs = [
    "max", "std", "cor",
]
labels = [
    "Max", "Std", "Cor",
]

args = dict(
    width = 600,
margin = dict(l=10, r=25, t=25, b=10),
pad=10,
)


"""
plot_data(paths, imgs, labels, dlabels, **args).write_image(
    "figs/data.pdf"
)
"""


plot_stats(paths, imgs, labels, dlabels, width=600, height=200).write_image(
    "figs/stats.pdf"
)



import numpy as np


def calc_sn(spikes):
    return np.array([si.max() / (1.4826 * np.median(si[si > 0])) for si in spikes])


def robust_zscore(x):
    medx = np.median(x)
    diff = x - medx
    stdx = 1.4826 * np.median(np.abs(x - medx))
    return diff / stdx


def calc_mah(x, y):
    dx = np.stack([x - x.mean(), y - y.mean()], axis=1)
    tau = np.linalg.inv((dx.T @ dx) / dx.shape[0])
    return np.sqrt(np.sum((dx @ tau) * dx, axis=1))



from logging import getLogger

import numpy as np
import pandas as pd

from ..footprint import clean
from ..spike import fix_kind
from ..io import (
    save,
    try_load,
)
from ..train import (
    SpatialModel,
    get_penalty,
)
from ..utils import (
    get_clip,
    get_xla_stats,
)
from .common import (
    get_data,
    get_files,
    get_force,
    rev_index,
)

logger = getLogger(__name__)


def spatial(cfg, stage, force=False):
    cleanstatsfile, footprintsfile = get_files(cfg, "clean", stage)
    if (
        get_force(cfg, "clean", stage)
        or not cleanstatsfile.exists()
        or not footprintsfile.exists
    ):
        stats = try_load(get_files(cfg, "evaluate", stage - 1))
        stats = stats.query("kind != 'remove'").copy()
        segstatsfile, segsfile, lossfile = get_files(cfg, "spatial", stage)
        if (
                get_force(cfg, "spatial", stage)
                or not segstatsfile.exists()
                or not segsfile.exists()
        ):
            logger.info(f"exec spatial ({stage})")
            data = get_data(cfg)
            if stage == 1:
                footprints = try_load(get_files(cfg, "make", stage - 1))
            else:
                _, footprints = try_load(get_files(cfg, "clean", stage - 1))
            spikes, bg, _ = try_load(get_files(cfg, "temporal", stage - 1))
            logger.debug("%s", get_xla_stats())
            stats, spikes, bg = fix_kind(
                stats,
                spikes,
                bg,
                cfg.dynamics,
                bg_type=cfg.bg_type,
                **(cfg.clean.temporal if stage > 1 else {}),
            )
            model = SpatialModel(
                data,
                stats,
                footprints,
                spikes,
                bg,
                cfg.dynamics,
                get_penalty(cfg.penalty, stage),
            )
            clips = get_clip(data.shape, cfg.cmd.spatial.clip)
            logdfs = []
            out = []
            for i, clip in enumerate(clips):
                model.prepare(clip, **cfg.cmd.spatial.prepare)
                log = model.fit(**cfg.cmd.spatial.step)
                logger.debug("%s", get_xla_stats())

                loss, sigma = zip(*log)
                df = pd.DataFrame(
                    dict(
                        stage="stage",
                        kind="spatial",
                        div=i,
                        step=np.arange(len(log)),
                        loss=loss,
                        sigma=sigma,
                    )
                )
                logdfs.append(df)
                out.append(model.get_x())
            logdf = pd.concat(logdfs, axis=0)
            index, x = (np.concatenate(v, axis=0) for v in zip(*out))
            logger.debug(
                "%d %d\n%s",
                index.size,
                np.count_nonzero(np.sort(index) != np.arange(index.size)),
                index,
            )

            segments = x[rev_index(index)]
            if cfg.fix_top:
                nk, h, w = segments.shape
                rseg = segments.reshape(nk, h * w)
                idx = np.argpartition(rseg, -2, axis=1)
                k = np.arange(nk)
                rseg[k, idx[:, -1]] = rseg[k, idx[:, -2]]
                segments = rseg.reshape(nk, h, w)
            stats["segid"] = np.arange(stats.shape[0])
            save((segstatsfile, segsfile, lossfile), (stats, segments, logdf))
            logger.info(f"saved spatial ({stage})")
        else:
            logger.info(f"load spatial ({stage})")
            segments = try_load(segsfile)
        logger.info(f"exec clean ({stage})")
        stats, footprints = clean(
            stats,
            segments,
            cfg.radius.filter,
            bg_type=cfg.bg_type,
            **cfg.clean.spatial,
            **cfg.cmd.clean,
        )
        save((cleanstatsfile, footprintsfile), (stats, footprints))
        logger.info(f"saved clean ({stage})")
        if cfg.cmd.remove_segments:
            segsfile.unlink(missing_ok=True)



import pandas as pd
import plotly.express as px
import plotly.io as pio

pio.kaleido.scope.mathjax = None


lus = [0, 20, 40, 100]
dfs = []
for lu in lus:
    path = f"CA3/miniature/min21/default/u{lu}"
    df = pd.read_csv(f"{path}/000stats.csv")
    df = df.query("kind != 'remove'")
    # fp = np.load(f"{path}/../../000footprint.npy")
    # df["intensity"] = df.signal * df.firmness
    df["lu"] = lu
    df["size"] = 1
    df["symbol"] = 1
    dfs.append(df)
df = pd.concat(dfs, axis=0)
fig = px.scatter(
    df,
    x="udense",
    y="signal",
    color=df.signal < 1,
    size="size",
    size_max=3,
    opacity=0.5,
    facet_col="lu",
)
for i in range(8):
    fig.data[i].marker.line.width = 0

for lu in lus:
    fig.update_annotations(selector=dict(text=f"lu={lu}"), y=0.8, text=f"λU={lu}")
fig.update_layout(
    template="none",
    font_size=11,
    showlegend=False,
    width=600,
    height=150,
    margin=dict(l=35, r=10, t=20, b=32),
)
fig.show()
fig.write_image("../figs/sig-dense.pdf")



from collections import namedtuple
from logging import getLogger

import jax
import jax.numpy as jnp
import numpy as np
import tensorflow as tf
from scipy.ndimage import grey_closing

from ..filter import gaussian_laplace
from ..utils import (
    from_tf,
    get_gpu_env,
)
from .radius import get_radius
from .segment import get_segment_mask

# from sklearn.mixture import GaussianMixture as CluModel


logger = getLogger(__name__)

Footprint = namedtuple("Footprint", "foootprit y x radius intensity")


def clean(
    stats,
    segs,
    radius,
    cell_range,
    thr_move,
    dupfilter=None,
    bgfilter=None,
    env=None,
    factor=1,
    prefetch=1,
):
    segments, y, x, radius, firmness = clean_footprints(
        segs,
        radius,
        env,
        factor,
        prefetch,
    )
    stats = stats.sort_values("segid")
    oldy = stats.y.to_numpy()
    oldx = stats.x.to_numpy()
    pos_move = np.hypot(x - oldx, y - oldy) / stats.radius.to_numpy()

    stats["old_kind"] = stats.kind
    stats["y"] = y
    stats["x"] = x
    stats["pos_move"] = pos_move
    stats["radius"] = radius
    stats["firmness"] = firmness
    stats["kind"] = ""
    stats["dup"] = -1

    dup_filter = get_dupfilter(**dupfilter)
    dup_filter.set(stats, segments)

    bg_filter = get_bgfilter(**bgfilter)
    bg_filter.set(stats)

    flg = np.argsort(stats.firmness.to_numpy())[::-1]
    cell = []
    bg = []
    while flg.size > 0:
        i, flg = flg[0], flg[1:]
        if (
            (radius[i] < cell_range[0])
            or ((stats.iloc[i].old_kind == "cell") and (pos_move[i] > thr_move))
        ):
            logger.debug("remove small/move %s %s %s", i, radius[i], pos_move[i])
            stats.at[stats.index[i], "kind"] = "remove"
        elif (
            (stats.iloc[i].old_kind == "background")
            or (radius[i] > cell_range[1])
            or bg_filter.is_background(i)
        ):
            if bg and ((dup := dup_filter.dup_id(i, bg)) >= 0):
                stats.at[stats.index[i], "kind"] = "remove"
                stats.at[stats.index[i], "dup"] = stats.index[dup]
            else:
                bg.append(i)
                stats.at[stats.index[i], "kind"] = "background"
        else:
            if cell and ((dup := dup_filter.dup_id(i, cell)) >= 0):
                stats.at[stats.index[i], "kind"] = "remove"
                stats.at[stats.index[i], "dup"] = stats.index[dup]
            else:
                cell.append(i)
                stats.at[stats.index[i], "kind"] = "cell"
    logger.info("cell/bg = %d/%d", len(cell), len(bg))
    return segments, stats


def clean_footprints(segs, radius, env=None, factor=1, prefetch=1):
    @jax.jit
    def calc(imgs):
        imgs /= imgs.max(axis=(1, 2), keepdims=True)
        gl = gaussian_laplace(imgs, radius, -3)
        nk, nr, h, w = gl.shape
        idx = jnp.argmax(gl.reshape(nk, nr * h * w), axis=1)
        k = jnp.arange(nk)
        r, y, x = idx // (h * w), (idx // w) % h, idx % w
        g = gl[k, r, y, x]
        seg = jax.vmap(get_segment_mask)(gl[k, r], y, x)
        imgs = jnp.where(seg, imgs, jnp.nan)
        dmin = jnp.nanmin(jnp.where(seg, imgs, jnp.nan), axis=(1, 2), keepdims=True)
        dmax = jnp.nanmax(jnp.where(seg, imgs, jnp.nan), axis=(1, 2), keepdims=True)
        imgs = jnp.where(seg, (imgs - dmin) / (dmax - dmin), 0)
        return imgs, y, x, r, g

    radius = get_radius(radius)
    nk, h, w = segs.shape

    env = get_gpu_env(env)
    nd = env.num_devices
    sharding = env.sharding((nd, 1))
    batch = env.batch(float(factor) * h * w * len(radius), nk)

    dataset = tf.data.Dataset.from_generator(
        lambda: zip(range(nk), segs),
        output_signature=(
            tf.TensorSpec((), tf.int32),
            tf.TensorSpec((h, w), tf.float32),
        ),
    )
    dataset = dataset.batch(batch)
    dataset = dataset.prefetch(prefetch)

    out = jnp.empty((nk + 1, h, w))
    y = jnp.empty((nk + 1,), jnp.int32)
    x = jnp.empty((nk + 1,), jnp.int32)
    r = jnp.empty((nk + 1,), jnp.int32)
    g = jnp.empty((nk + 1,), jnp.float32)

    logger.info("clean: %s %s", (factor, h, w), batch)
    logger.info("%s: %s %s %d", "pbar", "start", "clean", nk)
    for data in dataset:
        data = (from_tf(v) for v in data)
        idx, img = (jax.device_put(v, sharding) for v in data)

        count = idx.size
        diff = batch - count
        if diff > 0:
            pad = (0, diff), (0, 0), (0, 0)
            idx = jnp.pad(idx, pad[:1], constant_values=-1)
            img = jnp.pad(img, pad, constant_values=jnp.nan)

        outi, yi, xi, ri, gi = calc(img)
        out = out.at[idx].set(outi)
        y = y.at[idx].set(yi)
        x = x.at[idx].set(xi)
        r = r.at[idx].set(ri)
        g = g.at[idx].set(gi)
        logger.info("%s: %s %d", "pbar", "update", count)
    logger.info("%s: %s", "pbar", "close")

    out, y, x, r, g = (np.array(v[:-1]) for v in (out, y, x, r, g))
    out = grey_closing(out, (1, 10, 10))
    r = np.array(radius)[r]
    return Footprint(out, y, x, r, g)


def get_dupfilter(**args):
    match args:
        case {"thr_active_area": thr}:
            return DupFilter(thr)
        case _:
            raise ValueError()


def get_bgfilter(**args):
    match args:
        case {"kind": "new", **kwargs}:
            return BackgroundFilter(**kwargs)
        case {"kind": "simple", **kwargs}:
            return SimpleBackgroundFilter(**kwargs)
        case {"kind": "factor", **kwargs}:
            return FactorBackgroundFilter(**kwargs)
        case _:
            return OldBackgroundFilter(**args)


class DupFilter:
    def __init__(self, thr_active_area):
        self._thr = thr_active_area

    def set(self, stats, segments):
        self._y = stats.y.to_numpy()
        self._x = stats.x.to_numpy()
        self._segmask = segments > self._thr

    def dup_id(self, i, js):
        yi, xi = self._y[i], self._x[i]
        for j in js:
            yj, xj = self._y[j], self._x[j]
            if self._segmask[i, yj, xj] and self._segmask[j, yi, xi]:
                return j
        return -1


class BackgroundFilter:
    def __init__(
        self,
        bias=0,
        coef_log_snr=0,
        coef_firmness=0,
    ):
        self._bias = bias
        self._coef_log_snr = coef_log_snr
        self._coef_firmness = coef_firmness

    def set(self, stats):
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.iloc[i]
        score = (
            self._bias
            + self._coef_log_snr * np.log10(si.snratio)
            + self._coef_firmness * si.firmness
        )
        return score < 0


class FactorBackgroundFilter:
    def __init__(
        self,
        thr_rsn_factor=10,
        thr_firmness=0,
    ):
        self._thr_s = thr_rsn_factor
        self._thr_f = thr_firmness

    def set(self, stats):
        stats["rsn"] = 1 / stats.snratio
        med = np.median(stats.rsn)
        std = 1.4826 * np.median(np.abs(stats.rsn - med))
        stats["zrsn"] = (stats.rsn - med) / std
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.iloc[i]
        return (
            (si.zrsn > self._thr_s)
            or (si.firmness < self._thr_f)
        )


class SimpleBackgroundFilter:
    def __init__(
        self,
        thr_snr_inv=1,
        thr_firmness=0,
    ):
        self._thr_s = thr_snr_inv
        self._thr_f = thr_firmness

    def set(self, stats):
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.iloc[i]
        return (
            (1 / si.snratio > self._thr_s)
            or (si.firmness < self._thr_f)
        )


class OldBackgroundFilter:
    def __init__(
        self,
        thr_bg_udense=1,
        thr_bg_signal=0,
        thr_bg_firmness=0,
        thr_bg_cluster=1,
    ):
        self._thr_d = thr_bg_udense
        self._thr_s = thr_bg_signal
        self._thr_f = thr_bg_firmness
        self._thr_c = thr_bg_cluster

    def set(self, stats):
        """
        cindex = oldstats.query("kind == 'cell'").cid
        v = np.stack([udense, firmness, np.log(signal)], axis=1)
        v = v[cindex]
        clu = CluModel(
            n_components=2,
            weights_init=[0.95, 0.05],
            means_init=[[0.03, 0.45, 0.5], [0.13, 0.35, -0.3]],
        )
        clu.fit(v)
        z[cindex] = clu.predict_proba(v)[:, 1]
        self._stats.
        """
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.at[s.index[i]]
        return (
            # (si.z > thr_bg_cluster)
            (si.udense > self._thr_d)
            or (si.signal < self._thr_s)
            or (si.firmness < self._thr_f)
        )



from collections import namedtuple
from logging import getLogger

import jax
import jax.lax as lax
import jax.numpy as jnp
import numpy as np
import pandas  as pd
import tensorflow as tf
from scipy.ndimage import grey_closing

from ..filter import gaussian_laplace
from ..utils import (
    from_tf,
    get_gpu_env,
)
from ..utils.math import (
    calc_sn,
    robust_zscore,
    calc_mah,
)
from .radius import get_radius
from .segment import get_segment_mask

# from sklearn.mixture import GaussianMixture as CluModel


logger = getLogger(__name__)

Footprint = namedtuple("Footprint", "foootprit y x radius intensity")


def clean(
    stats,
    segs,
    radius,
    bg_type="bg",
    cell_range=None,
    thr_remove=None,
    thr_bg=None,
    thr_cell=None,
    env=None,
    factor=1,
    prefetch=1,
):
    segments, y, x, radius, firmness = clean_footprints(
        segs,
        radius,
        env,
        factor,
        prefetch,
    )

    ratio = []
    for ai in segments:
        gy, gx = np.where(ai > 0)
        w = ai[gy, gx]
        if w.size > 0:
            try:
                s = np.sqrt(np.linalg.eigvalsh(np.cov(gy, gx, aweights=w, ddof=0)))
                v = s.max() / s.min()
            except np.linalg.LinAlgError:
                v = np.nan
        else:
            v = np.nan
        ratio.append(v)
    ratio = np.array(ratio, np.float32)

    stats = stats.sort_values("segid")
    oldy = stats.y.to_numpy()
    oldx = stats.x.to_numpy()
    oldr = stats.radius.to_numpy()

    stats["old_kind"] = stats.kind
    stats["y"] = y
    stats["x"] = x
    stats["radius"] = radius
    stats["aratio"] = ratio
    stats["firmness"] = firmness
    stats["pos_move"] = np.hypot(x - oldx, y - oldy) / oldr

    cell_df = stats.query("old_kind == 'cell'")
    firmness = cell_df.firmness.to_numpy()
    rsn = cell_df.rsn.to_numpy()
    hypot = np.hypot(robust_zscore(firmness), robust_zscore(rsn))
    mah = calc_mah(firmness, rsn)
    stats["hypot"] = None
    stats.loc[cell_df.index, "hypot"] = hypot
    stats["mah"] = None
    stats.loc[cell_df.index, "mah"] = mah

    bg_cond = (stats.old_kind == "background")
    to_cell_cond = bg_cond.copy()
    for k, thr in thr_cell.items():
        to_cell_cond &= (stats[k] > thr)
    bg_cond &= ~to_cell_cond

    remove_cond = np.zeros(stats.shape[0], bool)
    for k, thr in thr_remove.items():
        if k == "pos_move":
            remove_cond |= (stats.old_kind == "cell") & (stats[k] > thr)
        elif k == "radius":
            remove_cond |= (stats[k] < thr)
        else:
            raise ValueError()

    tmp_cond = np.zeros(stats.shape[0], bool)
    for k, thr in thr_bg.items():
        tmp_cond |= (stats[k] > thr)
    if bg_type == "remove":
        remove_cond |= tmp_cond
    elif bg_type == "bg":
        bg_cond |= tmp_cond
    else:
        raise ValueError()

    bg_cond = bg_cond.to_numpy()
    remove_cond = remove_cond.to_numpy()

    stats["kind"] = "cell"
    stats.loc[bg_cond, "kind"] = "background"
    stats.loc[remove_cond, "kind"] = "remove"
    with pd.option_context("display.max_rows", None):
        print(stats[["kind", "old_kind", "segid", "radius", "pos_move", "aratio"]])

    cell_cond = (stats.kind == "cell").to_numpy()
    ys = y[cell_cond]
    xs = x[cell_cond]
    rs = radius[cell_cond]

    dist_mat = np.hypot(xs[:, np.newaxis] - xs, ys[:, np.newaxis] - ys) / rs
    np.fill_diagonal(dist_mat, np.inf)

    val_mat = segments[cell_cond][:, ys, xs]
    np.fill_diagonal(val_mat, 0)

    stats["max_dup_id"] = -1
    stats["min_dist_id"] = -1
    cell_df = stats.loc[cell_cond].copy()
    cell_df["min_dist_id"] = cell_df.index[np.argmin(dist_mat, axis=0)]
    cell_df["min_dist"] = dist_mat.min(axis=0)
    cell_df["max_dup_id"] = cell_df.index[np.argmax(val_mat, axis=0)]
    cell_df["max_dup"] = val_mat.max(axis=0)
    bg_df = stats.query("kind == 'background'").copy()
    remove_df = stats.query("kind == 'remove'").copy()
    stats = pd.concat([cell_df, bg_df, remove_df], axis=0)

    stats["spkid"] = -1
    stats["bgid"] = -1

    nc = cell_df.shape[0]
    nb = bg_df.shape[0]
    nr = remove_df.shape[0]
    logger.info("cell/bg/remove = %d/%d/%d", nc, nb, nr)
    return stats, segments


def clean_footprints(segs, radius, env=None, factor=1, prefetch=1):
    @jax.jit
    def calc(imgs):
        nk, h, w = imgs.shape
        k = jnp.arange(nk)
        imgs /= imgs.max(axis=(1, 2), keepdims=True)
        gl = gaussian_laplace(imgs, radius, -3)
        nk, nr, h, w = gl.shape
        idx = jnp.argmax(gl.reshape(nk, nr * h * w), axis=1)
        r, y, x = idx // (h * w), (idx // w) % h, idx % w
        g = gl[k, r, y, x]
        seg = jax.vmap(get_segment_mask)(gl[k, r], y, x)
        imgs = jnp.where(seg, imgs, 0)
        dmin = jnp.min(jnp.where(seg, imgs, jnp.inf), axis=(1, 2), keepdims=True)
        dmax = jnp.max(jnp.where(seg, imgs, -jnp.inf), axis=(1, 2), keepdims=True)
        imgs = jnp.where(seg, (imgs - dmin) / (dmax - dmin), 0)
        return imgs, y, x, r, g

    radius = get_radius(radius)
    nk, h, w = segs.shape

    env = get_gpu_env(env)
    nd = env.num_devices
    sharding = env.sharding((nd, 1))
    batch = env.batch(float(factor) * h * w * len(radius), nk)

    dataset = tf.data.Dataset.from_generator(
        lambda: zip(range(nk), segs),
        output_signature=(
            tf.TensorSpec((), tf.int32),
            tf.TensorSpec((h, w), tf.float32),
        ),
    )
    dataset = dataset.batch(batch)
    dataset = dataset.prefetch(prefetch)

    out = jnp.empty((nk + 1, h, w))
    y = jnp.empty((nk + 1,), jnp.int32)
    x = jnp.empty((nk + 1,), jnp.int32)
    r = jnp.empty((nk + 1,), jnp.int32)
    g = jnp.empty((nk + 1,), jnp.float32)

    logger.debug("clean: %s %s", (factor, h, w), batch)
    logger.info("%s: %s %s %d", "pbar", "start", "clean", nk)
    for data in dataset:
        data = (from_tf(v) for v in data)
        idx, img = (jax.device_put(v, sharding) for v in data)

        count = idx.size
        diff = batch - count
        if diff > 0:
            pad = (0, diff), (0, 0), (0, 0)
            idx = jnp.pad(idx, pad[:1], constant_values=-1)
            img = jnp.pad(img, pad, constant_values=jnp.nan)

        outi, yi, xi, ri, gi = calc(img)
        out = out.at[idx].set(outi)
        y = y.at[idx].set(yi)
        x = x.at[idx].set(xi)
        r = r.at[idx].set(ri)
        g = g.at[idx].set(gi)
        logger.info("%s: %s %d", "pbar", "update", count)
    logger.info("%s: %s", "pbar", "close")

    out, y, x, r, g = (np.array(v[:-1]) for v in (out, y, x, r, g))
    out = grey_closing(out, (1, 10, 10))
    r = np.array(radius)[r]
    return Footprint(out, y, x, r, g)


def get_dupfilter(**args):
    match args:
        case {"thr_active_area": thr}:
            return DupFilter(thr)
        case _:
            raise ValueError()


def get_bgfilter(**args):
    match args:
        case {"kind": "new", **kwargs}:
            return BackgroundFilter(**kwargs)
        case {"kind": "simple", **kwargs}:
            return SimpleBackgroundFilter(**kwargs)
        case {"kind": "factor", **kwargs}:
            return FactorBackgroundFilter(**kwargs)
        case _:
            return OldBackgroundFilter(**args)


class DupFilter:
    def __init__(self, thr_active_area):
        self._thr = thr_active_area

    def set(self, stats, segments):
        self._y = stats.y.to_numpy()
        self._x = stats.x.to_numpy()
        self._segmask = segments > self._thr

    def dup_id(self, i, js):
        yi, xi = self._y[i], self._x[i]
        for j in js:
            yj, xj = self._y[j], self._x[j]
            if self._segmask[i, yj, xj] and self._segmask[j, yi, xi]:
                return j
        return -1


class BackgroundFilter:
    def __init__(
        self,
        bias=0,
        coef_log_snr=0,
        coef_firmness=0,
    ):
        self._bias = bias
        self._coef_log_snr = coef_log_snr
        self._coef_firmness = coef_firmness

    def set(self, stats):
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.iloc[i]
        score = (
            self._bias
            + self._coef_log_snr * np.log10(si.snratio)
            + self._coef_firmness * si.firmness
        )
        return score < 0


class FactorBackgroundFilter:
    def __init__(
        self,
        thr_rsn_factor=10,
        thr_firmness=0,
    ):
        self._thr_s = thr_rsn_factor
        self._thr_f = thr_firmness

    def set(self, stats):
        stats["rsn"] = 1 / stats.snratio
        med = np.median(stats.rsn)
        std = 1.4826 * np.median(np.abs(stats.rsn - med))
        stats["zrsn"] = (stats.rsn - med) / std
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.iloc[i]
        return (
            (si.zrsn > self._thr_s)
            or (si.firmness < self._thr_f)
        )


class SimpleBackgroundFilter:
    def __init__(
        self,
        thr_snr_inv=1,
        thr_firmness=0,
    ):
        self._thr_s = thr_snr_inv
        self._thr_f = thr_firmness

    def set(self, stats):
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.iloc[i]
        return (
            (1 / si.snratio > self._thr_s)
            or (si.firmness < self._thr_f)
        )


class OldBackgroundFilter:
    def __init__(
        self,
        thr_bg_udense=1,
        thr_bg_signal=0,
        thr_bg_firmness=0,
        thr_bg_cluster=1,
    ):
        self._thr_d = thr_bg_udense
        self._thr_s = thr_bg_signal
        self._thr_f = thr_bg_firmness
        self._thr_c = thr_bg_cluster

    def set(self, stats):
        """
        cindex = oldstats.query("kind == 'cell'").cid
        v = np.stack([udense, firmness, np.log(signal)], axis=1)
        v = v[cindex]
        clu = CluModel(
            n_components=2,
            weights_init=[0.95, 0.05],
            means_init=[[0.03, 0.45, 0.5], [0.13, 0.35, -0.3]],
        )
        clu.fit(v)
        z[cindex] = clu.predict_proba(v)[:, 1]
        self._stats.
        """
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.at[s.index[i]]
        return (
            # (si.z > thr_bg_cluster)
            (si.udense > self._thr_d)
            or (si.signal < self._thr_s)
            or (si.firmness < self._thr_f)
        )



from tqdm import tqdm


class Progress:
    def __init__(self, setter):
        self._setter = setter

    def skip(self):
        pass

    def session(self, name):
        print("session", name)
        self._name = name
        self._setter((self._name, "", "", ""))
        return self

    def set_count(self, total, status=""):
        print("set_count", total, status)
        self._status = status
        self._total = total
        self._n = 0
        self._setter((self._name, str(self._n), str(self._total), str(self._status)))

    def update(self, n, status=""):
        print("update", n, status)
        self._status = status
        self._n += n
        self._setter((self._name, str(self._n), str(self._total), str(self._status)))


class SilentProgress:
    def skip(self):
        pass

    def session(self, name):
        return self

    def set_count(self, total, status=None):
        pass

    def update(self, n, status=None):
        pass


class TQDMProgress:
    def __init__(self):
        self._curr = None

    def skip(self):
        pass

    def session(self, name):
        # print("session", name)
        if self._curr is not None:
            self._curr.close()
        self._name = name
        return self

    def set_count(self, total, status=None):
        # print("set_count", total, status)
        self._curr = tqdm(desc=self._name, total=total)

    def update(self, n, status=None):
        # print("update", n, status)
        if status is not None:
            self._curr.set_postfix_str(status, refresh=False)
        self._curr.update(n)



import dash.dcc as dcc
import dash.html as html
import dash_bootstrap_components as dbc
import numpy as np
from dash import (
    Input,
    Output,
    State,
    callback,
)
from omegaconf import OmegaConf


def InputGroup(*components, **kwargs):
    style = kwargs.setdefault("style", {})
    style.update(
        display="grid",
        gridTemplateColumns="auto 1fr",
    )
    return html.Div(
        children=sum([[dbc.Label(k), v] for k, v in components], []),
        **kwargs,
    )


def create_config_div(cfg):
    def Radius1Slider(value, rrange, skip=10, **kwargs):
        return dcc.RangeSlider(
            min=0,
            max=len(rrange) - 1,
            value=[np.searchsorted(rrange, v) for v in value],
            marks={i: f"{rrange[i]:.2f}" for i in range(0, len(rrange), skip)},
            **kwargs,
        )

    radius_range = np.geomspace(2, 32, 41)

    imgs_path = dbc.Input(type="text", value=cfg.data.imgs.file)
    imgs_mask = dbc.Input(type="text", value=cfg.data.mask.type)
    imgs_hz = dbc.Input(type="number", value=cfg.data.hz)
    radius1 = Radius1Slider([cfg.radius.rmin, cfg.radius.rmax], radius_range)
    nradius = dcc.Slider(min=5, max=30, step=1, value=cfg.radius.rnum)
    radius2 = dcc.RangeSlider(
        min=0,
        max=cfg.radius.rnum - 1,
        step=1,
        value=[cfg.radius.rmin2, cfg.radius.rmax2],
    )
    thr = dcc.Slider(id="thr", min=0.5, max=3, step=0.1, value=cfg.radius.thr)
    tau = dcc.RangeSlider(
        min=0,
        max=500,
        step=10,
        value=[1000 * cfg.dynamics.tau1, 1000 * cfg.dynamics.tau2],
        marks={i: str(i) for i in range(0, 501, 100)},
    )
    duration = dcc.Slider(min=0.0, max=3.0, value=cfg.dynamics.duration)
    la = dcc.Slider(min=0.0, max=100.0, value=cfg.penalty.la)
    lu = dcc.Slider(min=0.0, max=100.0, value=cfg.penalty.lu)
    bx = dcc.Slider(min=0.0, max=1.0, value=cfg.penalty.bx)
    bt = dcc.Slider(min=0.0, max=1.0, value=cfg.penalty.bt)

    div = html.Div(
        children=[
            cfg_store := dcc.Store("cfg_store", data=OmegaConf.to_container(cfg)),
            InputGroup(
                ("Image file path", imgs_path),
                ("Mask type", imgs_mask),
                ("Frequency", imgs_hz),
                style=dict(width="500px"),
            ),
            InputGroup(
                ("Radius range (px)", radius1),
                ("Num radius", nradius),
                ("Radius cut range", radius2),
                ("Distance threshold", thr),
                ("Time constants (msec)", tau),
                ("Duration (sec)", duration),
                ("Footprint sparseness", la),
                ("Spike sparseness", lu),
                ("Spatial background", bx),
                ("Temporal background", bt),
                style=dict(width="1200px"),
            ),
        ],
    )

    @callback(
        Output(cfg_store, "data"),
        Input(imgs_path, "value"),
        Input(imgs_mask, "value"),
        Input(imgs_hz, "value"),
        Input(radius1, "value"),
        Input(nradius, "value"),
        Input(radius2, "value"),
        Input(thr, "value"),
        Input(tau, "value"),
        Input(duration, "value"),
        Input(la, "value"),
        Input(lu, "value"),
        Input(bx, "value"),
        Input(bt, "value"),
        State(cfg_store, "data"),
    )
    def set_cfg(
        imgs, mask, hz, radius1, rnum, radius2, thr, tau, duration, la, lu, bx, bt, cfg
    ):
        rmin1, rmax1 = [radius_range[r] for r in radius1]
        rmin2, rmax2 = radius2
        tau1, tau2 = tau
        cfg = OmegaConf.create(cfg)
        cfg.data.imgs.file = imgs
        cfg.data.mask.type = mask
        cfg.data.hz = hz
        cfg.radius.rmin = float(rmin1)
        cfg.radius.rmax = float(rmax1)
        cfg.radius.rnum = rnum
        cfg.radius.rmin2 = rmin2
        cfg.radius.rmax2 = rmax2
        cfg.radius.thr = thr
        cfg.dynamics.tau1 = tau1 / 1000
        cfg.dynamics.tau2 = tau2 / 1000
        cfg.dynamics.duration = duration
        cfg.dynamics.duration = duration
        cfg.penalty.la = la
        cfg.penalty.lu = lu
        cfg.penalty.bx = bx
        cfg.penalty.bt = bt
        return OmegaConf.to_container(cfg)

    @callback(
        Output(radius2, "max"),
        Output(radius2, "marks"),
        Input(radius1, "value"),
        Input(nradius, "value"),
    )
    def set_radius2(radius1, rnum):
        rmin1, rmax1 = radius1
        rmin1 = radius_range[rmin1]
        rmax1 = radius_range[rmax1]
        radius = np.geomspace(rmin1, rmax1, rnum)
        marks = {i: f"{r:.2f}" for i, r in enumerate(radius)}
        return rnum - 1, marks

    div.store = cfg_store
    return div



import sys
from pathlib import Path
from importlib import import_module
from multiprocessing import Process

from ..plot import (
    peak_stats_fig,
    seg_max_fig,
)
from .common import set_env


def call(name, *args, **kwargs):
    def wrap(cfg, *args, **kwargs):
        set_env(cfg)
        return target(cfg, *args, **kwargs)

    target = getattr(import_module(f"hotaru.cui.{name}"), name)
    p = Process(target=wrap, args=args, kwargs=kwargs)
    p.start()
    p.join()
    if p.exitcode != 0:
        sys.exit()


def test(cfg):
    call("normalize", cfg)
    call("init", cfg)

    path = Path(cfg.outputs.figs.dir)
    path.mkdir(parents=True, exist_ok=True)
    peak_stats_fig(cfg, 0).write_image(path / "test_stats.pdf")
    seg_max_fig(cfg, 0).write_image(path / "test_footprints.pdf")



from .image import (
    load_imgs,
    apply_mask,
)
from .saver import (
    save,
    try_load,
)
from .movie import (
    gen_result_movie,
)

__all__ = [
    "load_imgs",
    "apply_mask",
    "save",
    "try_load",
    "gen_result_movie",
]



from collections import namedtuple
from logging import getLogger

import numpy as np

logger = getLogger(__name__)


class Data(namedtuple("Data", "imgs mask hz avgx avgt std0 min0 max0 min1 max1")):

    @property
    def nt(self):
        return self.imgs.shape[0]

    @property
    def ns(self):
        match self.mask:
            case None:
                return self.shape[0] * self.shape[1]
            case _:
                return np.count_nonzero(self.mask)

    @property
    def shape(self):
        return self.imgs.shape[1:]

    def clip(self, clip):
        return self._replace(
            imgs=clip(self.imgs),
            mask=clip(self.mask),
            avgx=clip(self.avgx),
        )

    def apply_mask(self, x, mask_type=np.nan):
        match (mask_type, self.mask):
            case (False, _):
                pass
            case (True, None):
                *shape, h, w = x.shape
                x = x.reshape(shape + [h * w])
            case (True, mask):
                x = x[..., mask]
            case (val, None):
                pass
            case (val, mask):
                x[..., mask] = val
            case _:
                raise ValueError()
        return x

    def select(self, s, mask_type=np.nan):
        imgs = self.imgs[s]
        avgx = self.avgx
        avgt = self.avgt[s, np.newaxis, np.newaxis]
        std0 = self.std0
        imgs = (imgs - avgx - avgt) / std0
        return self.apply_mask(imgs, mask_type)

    def data(self, idx=None, mask_type=np.nan):
        if idx is None:
            idx = range(self.nt)
        for i in idx:
            yield self.select(i, mask_type)

    def datax(self):
        data = self.imgs.reshape(self.nt, -1)
        avgx = self.avgx.ravel()
        avgt = self.avgt
        std0 = self.std0
        if self.mask is None:
            for d, a in zip(data.T, avgx):
                yield (d - a - avgt) / std0
        else:
            mask = self.mask.ravel()
            for m, d, a in zip(mask, data.T, avgx):
                if m:
                    yield (d - a - avgt) / std0



from functools import partial

import jax
import jax.lax as lax
import jax.numpy as jnp
import numpy as np


@partial(jax.jit, static_argnames=["rs", "axis"])
def gaussian_laplace(imgs, rs, axis=-1):
    ndim = imgs.ndim
    if axis < 0:
        axis = ndim + axis + 1
    out_shape = imgs.shape[:axis] + (len(rs),) + imgs.shape[axis:]
    out = jnp.empty(out_shape)
    for i, r in enumerate(rs):
        idx = (slice(None),) * axis + (i,) + (slice(None),) * (ndim - axis)
        out = out.at[idx].set(gaussian_laplace_single(imgs, r))
    return out


@partial(jax.jit, static_argnames=["r"])
def gaussian_laplace_single(imgs, r):
    return _gaussian_laplace(imgs, r, 4 * np.ceil(r))


def _gaussian_laplace(imgs, r, nd):
    sqrt_2pi = jnp.sqrt(2 * jnp.pi)
    d = jnp.square(jnp.arange(-nd, nd + 1, 1))
    r2 = jnp.square(r)
    o0 = jnp.exp(-d / r2 / 2) / r / sqrt_2pi
    o2 = (1 - d / r2) * o0
    o0 = jnp.expand_dims(o0, (0, 1))
    o2 = jnp.expand_dims(o2, (0, 1))
    gl1 = lax.conv(jnp.expand_dims(imgs, -3), jnp.expand_dims(o2, -1), (1, 1), "same")
    gl1 = lax.conv(gl1, jnp.expand_dims(o0, -2), (1, 1), "same")
    gl2 = lax.conv(jnp.expand_dims(imgs, -3), jnp.expand_dims(o2, -2), (1, 1), "same")
    gl2 = lax.conv(gl2, jnp.expand_dims(o0, -1), (1, 1), "same")
    return (gl1 + gl2)[..., 0, :, :]



import time

from logging import getLogger

default_logger = getLogger(__name__)


class Timer:
    def __init__(self, name=None, logger=None):
        self.name = name
        self.logger = logger

    def __enter__(self):
        self.time = time.time()
        self.ptime = time.process_time()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.time = time.time() - self.time
        self.ptime = time.process_time() - self.ptime
        if self.logger:
            logger = self.logger
        else:
            logger = default_logger
        logger.debug("%s: %f %f", self.name, self.time, self.ptime)

    def get(self):
        return self.time, self.ptime



import plotly.graph_objects as go
import pandas as pd
import numpy as np

from ..cui.common import load
from .common import add_jitter


def compare_lu_fig(paths, labels):
    fig = go.Figure().set_subplots(
        2,
        len(paths),
        row_heights=(1, 5),
        horizontal_spacing=0.05,
        vertical_spacing=0.01,
        #shared_xaxes="columns",
        #shared_yaxes="rows",
    )
    for i, path in enumerate(paths):
        stats = pd.read_csv(path, index_col=0)
        df = stats.query("kind != 'remove'").copy()
        df = add_jitter(df)
        fig.add_trace(
            go.Histogram(x=df.udense, marker_color="grey"),
            row=1,
            col=i + 1,
        )
        fig.add_trace(
            go.Scatter(
                y=df.intensity,
                x=df.udense,
                mode="markers",
                marker=dict(size=3, color="blue", opacity=0.5),
            ),
            row=2,
            col=i + 1,
        )
        fig.add_annotation(
            x=0.2,
            y=15,
            text=labels[i],
            showarrow=False,
            font_size=11,
            row=2,
            col=i + 1,
        )
        fig.update_xaxes(
            showticklabels=False,
            row=1,
            col=i + 1,
        )
        fig.update_yaxes(
            showticklabels=False,
            row=1,
            col=i + 1,
        )
        fig.update_xaxes(
            range=(0, 0.28),
            tickmode="array",
            tickvals=[0, 0.1, 0.2],
            ticktext=[0, 0.1, 0.2],
            title_text="spike density",
            row=2,
            col=i + 1,
        )
        fig.update_yaxes(
            title_text="intensity" if i ==0 else "",
            showticklabels=i == 0,
            #range=((4 / 5), (21 / 5)),
            #tickmode="array",
            #tickvals=np.log2([3, 6, 12]),
            #ticktext=["3", "6", "12"],
            row=2,
            col=i + 1,
        )
    fig.update_layout(
        template="none",
        font_size=11,
        width=600,
        height=200,
        showlegend=False,
        margin=dict(t=0, r=10, l=40, b=40),
    )
    return fig



import jax.numpy as jnp


class Regularizer:
    def __call__(self, x):
        return 0

    def prox(self, y, eta):
        return y


class L2:
    def __call__(self, x, fac):
        return jnp.square(fac * x).sum()

    def prox(self, y, eta, fac):
        return y / (1 + fac * eta)


class NonNegativeL2:
    def __call__(self, x, fac):
        return jnp.square(fac * x).sum()

    def prox(self, y, eta, fac):
        return jnp.maximum(0, y) / (1 + fac * eta)


class L1:
    def __call__(self, x, fac):
        return jnp.abs(fac * x).sum()

    def prox(self, y, eta, fac):
        return jnp.sign(y) * jnp.maximum(0, jnp.abs(y) - fac * eta)


class NonNegativeL1:
    def __call__(self, x, fac):
        x = jnp.maximum(0, x)
        return (fac * x).sum()

    def prox(self, y, eta, fac):
        return jnp.maximum(0, y - fac * eta)


class MaxNormNonNegativeL1:
    def __call__(self, x, fac):
        x = jnp.maximum(0, x)
        m = x.max(axis=-1)
        s = (fac * x).sum(axis=-1)
        if x.ndim > 1:
            positive = m > 0
            m = jnp.where(positive, m, 1)
            s = jnp.where(positive, s, 0)
            return (s / m).sum()
        else:
            if m > 0:
                return s / m
            else:
                return jnp.zeros(())

    def prox(self, y, eta, fac):
        y = jnp.maximum(0, y)
        m = y.max(axis=-1, keepdims=True)
        ym = jnp.where(y == m, y, y - fac * eta / m)
        return jnp.maximum(0, ym)



import dash.html as html
import dash_bootstrap_components as dbc
import diskcache
from dash import (
    Dash,
    DiskcacheManager,
)

from .config import create_config_div
from .init import create_init_div
from .peaks import create_peaks_div
from .stats import create_stats_div

"""
from ..footprint.find import simple_peaks
from ..main import (
    data,
    find,
    footprint,
    get_frame,
    make,
    reduce,
    spike,
    stats,
)
from .fig import (
    bar_fig,
    circle_fig,
    heat_fig,
    scatter_fig,
    spike_fig,
    update_img,
)
from .ui import (
    Collapse,
    ThreadButton,
    two_column,
)
"""


def gui(cfg, *args, **kwargs):
    kwargs.setdefault("external_stylesheets", [dbc.themes.BOOTSTRAP])
    kwargs.setdefault("title", "HOTARU")

    cache = diskcache.Cache(f"{cfg.outdir}/server_cache")
    manager = DiskcacheManager(cache)
    app = Dash(__name__, *args, **kwargs, background_callback_manager=manager)

    app.layout = html.Div(
        children=[
            cfg_div := create_config_div(cfg),
            dbc.Tabs(
                children=[
                    dbc.Tab(label="STATS", children=create_stats_div(cfg_div.store)),
                    dbc.Tab(label="PEAKS", children=create_peaks_div(cfg_div.store)),
                    dbc.Tab(label="INIT", children=create_init_div(cfg_div.store)),
                ],
            ),
        ]
    )

    return app



from logging import getLogger

import jax
import jax.numpy as jnp
import av
import matplotlib.pyplot as plt
import numpy as np
from PIL import (
    Image,
    ImageDraw,
    ImageFont,
)

logger = getLogger(__name__)


def gen_movie(filename, imgs, shape, fps, **kwargs):
    logger.info("%s: %s %s %d", "pbar", "start", "movie", shape[0])
    with av.open(filename, "w") as output:
        #codec = av.CodecContext.create("h264_nvenc", "w")
        #codec = av.CodecContext.create("libx264", "w")
        stream = output.add_stream(kwargs.get("codec", "h264"), fps)
        #stream = output.add_stream(codec, fps)
        stream.bit_rate = kwargs.get("bit_rate", kwargs.get("bit_rate", 8_000_000))
        stream.pix_fmt = kwargs.get("fmt", kwargs.get("fmt", "yuv420p"))
        stream.height = shape[1]
        stream.width = shape[2]
        for img in imgs:
            frame = av.VideoFrame.from_ndarray(img, format="rgba")
            packet = stream.encode(frame)
            output.mux(packet)
            logger.info("%s: %s %d", "pbar", "update", 1)
        logger.info("%s: %s", "pbar", "close")


def gen_normalize_movie(outfile, data, **kwargs):
    imgs, mask, hz, avgx, avgt, std0, min0, max0, min1, max1 = data
    nt, h, w = imgs.shape

    def gen_frame():
        for t in range(nt):
            img0 = (imgs[t] - min0) / (max0 - min0)
            img1 = (imgs[t] - avgx - avgt[t]) / std0
            img1 = (img1 - min1) / (max1 - min1)
            img = np.concatenate([img0, img1], axis=1)
            img = (255 * plt.get_cmap("Greens")(img)).astype(np.uint8)
            if mask is not None:
                img[~mask] = 0
            yield (255 * img).astype(np.uint8)

    gen_movie(outfile, gen_frame(), (nt, h, 2 * w), hz, **kwargs)


def gen_result_movie(
    outfile, data, footprints, spikes, dynamics, t0=None, t1=None, **kwargs
):
    calcium = np.array(dynamics(spikes))
    #calcium -= calcium.min(axis=1, keepdims=True)
    #calcium -= np.median(calcium, axis=1, keepdims=True)
    #calcium = np.where(calcium > 0, calcium, 0.0)
    calcium /= calcium.max()
    if t0 is None:
        t0 = 0
    if t1 is None:
        t1 = calcium.shape[1]
    cmap = plt.get_cmap("Greens")

    dmin = jnp.array(data.min1)
    dmax = jnp.array(data.max1)
    calcium = jnp.array(calcium)
    footprints = jnp.array(footprints)

    def gen(data, t):
        img0 = (data - dmin) / (dmax - dmin)
        img1 = (calcium[:, t, jnp.newaxis, jnp.newaxis] * footprints).sum(axis=0)
        img1 = jnp.where(img1 < 1.0, img1, 1.0)

        img = np.concatenate([img0, img1], axis=1)
        img = (255 * cmap(img)).astype(np.uint8)
        return img

    def gen_frame():
        for t in range(t0, t1):
            img = np.array(gen(jnp.array(data.select(t)), jnp.array(t)))
            img = Image.fromarray(img)
            draw = ImageDraw.Draw(img)
            draw.text((10, 10), "Normalized original movie", "red", font=font)
            draw.text((w + 10, 10), "Reconstructed movie", "red", font=font)

            yield np.array(img)

    font = ImageFont.truetype(
        "/usr/share/fonts/truetype/ttf-bitstream-vera/VeraBd.ttf", 24
    )
    nk, _ = calcium.shape
    _, h, w = footprints.shape
    footprints = footprints[:nk]
    gen_movie(outfile, gen_frame(), (t1 - t0, h, 2 * w), data.hz, **kwargs)



import os
from logging import getLogger
from pathlib import Path

import numpy as np
import pandas as pd
from hydra.core.hydra_config import HydraConfig

from ..io import (
    apply_mask,
    load_imgs,
    try_load,
)
from .data import Data

logger = getLogger(__name__)


def set_env(cfg):
    hcfg = HydraConfig.get()
    for k, v in hcfg.job.env_set.items():
        os.environ[k] = v


def get_force(cfg, name, stage):
    force_dict = dict(
        normalize=[
            "normalize",
            "find",
            "reduce",
            "make",
            "init",
            "temporal",
            "evaluate",
        ],
        find=["find", "reduce", "make", "init", "temporal", "evaluate"],
        reduce=["reduce", "make", "init", "temporal", "evaluate"],
        make=["make", "init", "temporal", "evaluate"],
        init=["init", "temporal", "evaluate"],
        spatial=["spatial", "clean", "temporal", "evaluate"],
        clean=["clean", "temporal", "evaluate"],
        temporal=["temporal", "evaluate"],
        evaluate=["evaluate"],
    )
    force = (stage > cfg.force_from[0]) or (
        (stage == cfg.force_from[0]) and (name in force_dict[cfg.force_from[1]])
    )
    logger.debug("test force: %s %d", name, stage)
    return force


def all_stats(cfg):
    out = []
    for stage in range(1000):
        stats, flag = load(cfg, "evaluate", stage)
        if flag is None:
            break
        out.append(stats)
    return out


def get_files(cfg, name, stage):
    odir = Path(cfg.outputs.dir)
    path = cfg.outputs[name]
    if (stage == 0) or ((stage == 1) and (name == "spatial")):
        fdir = odir / path.dir0
    else:
        fdir = odir / path.dir
    files = [fdir / file.format(stage=stage) for file in path.files]
    return files


def load(cfg, name, stage):
    return try_load(get_files(cfg, name, stage))


def get_data(cfg):
    imgs, hz = load_imgs(**cfg.data.imgs)
    imgs, mask = apply_mask(imgs, **cfg.data.mask)
    stats = try_load(get_files(cfg, "normalize", 0)[0])
    data = Data(imgs, mask, hz, *stats)
    return data


def rev_index(index):
    rev_index = np.full_like(index, index.size)
    for i, j in enumerate(index):
        rev_index[j] = i
    return rev_index


def reduce_log(cfg, stage):
    if stage == 0:
        stats = try_load(get_files(cfg, "reduce", stage))
    else:
        stats, _ = try_load(get_files(cfg, "clean", stage))
    cell = stats[stats.kind == "cell"]
    bg = stats[stats.kind == "background"]
    removed = stats[stats.kind == "remove"]

    radius = np.sort(np.unique(stats.radius))
    for r in radius:
        logger.info(
            "radius=%f cell=%d bg=%d remove=%d",
            r,
            (cell.radius == r).sum(),
            (bg.radius == r).sum(),
            (removed.radius == r).sum(),
        )


def print_stats(cfg, stage):
    stats = try_load(get_files(cfg, "evaluate", stage))

    cell = stats[stats.kind == "cell"]
    rsn = cell.rsn
    med = np.median(rsn)
    std = 1.4826 * np.median(np.abs(rsn - med))
    thr = med + 2 * std
    logger.info("rsn: med=%.3f std=%.3f thr=%.3f", med, std, thr)
    labels = [
        "y",
        "x",
        "radius",
        "firmness",
        "signal",
        # "udense",
        "unz",
        "rsn",
        "zrsn",
    ]
    if stage > 0:
        labels += [
            "min_dist_id",
            "min_dist",
            "max_dup_id",
            "max_dup",
        ]
    with pd.option_context("display.max_rows", None):
        logger.info(
            "cell: %d\n%s",
            cell.shape[0],
            cell[labels].sort_values("signal", ascending=False),
        )

    bg = stats[stats.kind == "background"]
    if bg.shape[0] > 0:
        labels = [
            "y",
            "x",
            "radius",
            "firmness",
            "bmax",
            "bsparse",
        ]
        # if stage > 0:
        #    labels += ["old_radius", "old_udense", "old_bsparse"]
        with pd.option_context("display.max_rows", None):
            logger.info(
                "background: %d\n%s",
                bg.shape[0],
                bg[labels].sort_values("bmax", ascending=False),
            )

    removed = stats[stats.kind == "remove"]
    if removed.shape[0] > 0:
        labels = [
            "y",
            "x",
            "radius",
            "firmness",
        ]
        if stage > 0:
            labels += [
                "pos_move",
            ]
        with pd.option_context("display.max_rows", None):
            logger.info(
                "removed: %d\n%s",
                removed.shape[0],
                removed[labels].sort_values("firmness", ascending=False),
            )


def finish(cfg, stage):
    stats = try_load(get_files(cfg, "evaluate", stage))
    removed = stats[stats.kind == "remove"]

    #reduce_log(cfg, stage)
    return cfg.early_stop and (stage > 0) and removed.shape[0] == 0



import logging

from tqdm import tqdm


class StreamHandler(logging.StreamHandler):

    def emit(self, record):
        try:
            match record.args:
                case ("pbar", "start", desc, total):
                    self.tqdm = tqdm(desc=desc, total=total, file=self.stream)
                case ("pbar", "start", desc, total, postfix):
                    self.tqdm = tqdm(desc=desc, total=total, file=self.stream)
                    self.tqdm.set_postfix_str(postfix)
                case ("pbar", "update", n):
                    self.tqdm.update(n)
                case ("pbar", "update", n, postfix):
                    self.tqdm.set_postfix_str(postfix, refresh=False)
                    self.tqdm.update(n)
                case ("pbar", "close"):
                    self.tqdm.close()
                case _:
                    msg = self.format(record)
                    tqdm.write(msg, end=self.terminator, file=self.stream)
                    self.flush()
        except RecursionError:
            raise
        except Exception:
            self.handleError(record)


class DropPbarFilter(logging.Filter):

    def filter(self, record):
        return (len(record.args) == 0) or (record.args[0] != "pbar")



from collections import namedtuple


class Stats(namedtuple("Stats", "t0 y0 x0 mask avgx avgt std0")):
    def normalize(self, imgs):
        return (self.clip(imgs) - self.avgx - self.avgt[:, None, None]) / self.std0

    def slice(self, start, end, batch):
        return self._replace(avgt=self.avgt[start:end].reshape(batch, -1))

    def clip(self, imgs):
        y0, x0 = self.y0, self.x0
        h, w = self.mask.shape
        return imgs[:, y0 : y0 + h, x0 : x0 + w]



import numpy as np
from plotly.colors import get_colorscale
from matplotlib.pyplot import get_cmap


def cmap(colorscale, value):
    mask = np.zeros_like(value, bool)
    out = np.empty_like(value, str)
    for thr, color in get_colorscale(colorscale):
        cond = mask & (value < thr)
        out[cond] = color
        mask |= cond
    return out


def to_image(data, cmap):
    return (255 * get_cmap(cmap)(data)).astype(np.uint8)


def add_jitter(df):
    radius = np.sort(np.unique(df.radius))
    rscale = np.log(np.min(radius[1:] / radius[:-1]))
    df["ri"] = df.radius * np.exp(rscale * (np.random.uniform(size=df.shape[0]) - 0.5))
    df["lri"] = np.log2(df.ri)
    return df
